A small program to experiment with `malloc`, useful in combination
with `pwndbg`.

To compile, just run `make`. You will obtain two binaries:

* `test-malloc`, which uses glibc's `malloc` implementation;

* `test-dlmalloc`, which uses Doug Lea's `malloc` (slightly changed to be
   compatible with `pwndbg`).

Run either program from `gdb`, with `pwndbg` loaded. The program prints a
prompt and waits for command lines. The program keeps an array of pointers to
chunks and lets you allocate/free/write into them.  At any time you can type
Ctrl+C and go back to the `gdb` prompt.  From there you can inspect the state
of the heap with, e.g., `vis` or `bins`. Just type `continue` (or simply `c`)
to resume the program execution where you can issue more commands.

The commands understood by the program are the following:

* `malloc size`: call `malloc(size)` and store the resulting
  pointer in the next available slot in the array. It prints
  the selected index in the array between square brackets,
  followed by the (user) address of the allocated chunk.
  The size can be written in base 10 (default), base 16
  (prefix it with `0x`) and base 8 (begin with `0`).

* `list`: print all the non-empty slots of the array.

* `free index`: call `free()` on the pointer stored at index `index`.
  This command does not clear the `index` slot, so you can experiment
  with double-free's and use-after-free's.

* `clear index`: writen `NULL` in slot `index` of the array.

* `writestr index string`: write `string` (whitespace delimited)
  into the user memory of the chunk at `index`. No bounds are checked,
  so you can experiment with overflows.

* `writehex index hexstring(s)`: write arbitrary bytes in the
  user memory of the chunk at `index`. Each byte must be written
  in hex, e.g.

	writehex 2 ef be ad de

  will write 0xdeadbeef into chunk 2 (spaces are optional).
  No bounds are checked.

Just type Ctrl+D on an empty line to exit.


Note
~~~~

It may be useful to send the output of the program to another terminal,
instead of sharing the same terminal for both the program and gdb.  Do as
follows:

1. run `gdb` in a first terminal, let's call it terminal 1;

2. open another terminal, let's call it terminal 2;

3. in terminal 2 enter `tty` and take note of the output
  (it's the device file of terminal 2);

4. in terminal 2 enter the following:
```sh
      clear; tail -f /dev/null
```
  The purpose of these commands it to clear the terminal and put to sleep
  all the processses that were using it (the shell will spawn `tail` and wait for it to
  terminate, while `tail` will wait for `/dev/null` to increase its size, which
  is not going to happen);

5. in terminal 1, where `gdb` is running, enter the command `tty` followed by
  the device file found in step 3.

6. now run the program in `gdb`. The program will use terminal 2 for input/output.
  When you want to inspect the program state, type Ctr+C in _terminal 1_.
