#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define MAX_SLOTS 100
#define BUF_SIZE 128

#ifdef USE_DLMALLOC
#define PWNDBG_COMPAT 1
#include "malloc-2.7.2.c"
#endif

void *slot[MAX_SLOTS];

static int interactive = 0;

void dummy()
{
	printf("called dummy\n");
}

void dummy2(const char *p)
{
	printf("called dummy2(\"%s\")\n", p);
}

int findfree()
{
	int i;

	for (i = 0; i < MAX_SLOTS; i++)
		if (!slot[i])
			return i;
	return -1;
}

void drain()
{
	int c;

	c = 0;
	while (!ferror(stdin) && !feof(stdin) && c != '\n')
		c = fgetc(stdin);
	if (interactive)
		clearerr(stdin);
}

int checkeol(const char *p)
{
	int c;

	if (!p)
		return 1;
	while (*p != '\0' && isspace(*p))
		p++;
	if (*p != '\0')
		drain();
	return 1;
}

int getslot()
{
	char *arg;
	long v;

	if ( !(arg = strtok(NULL, " \t")) ) {
		fprintf(stderr, "missing slot index\n");
		return -1;
	}
	v = strtol(arg, &arg, 0);
	if (v < 0 || v > MAX_SLOTS) {
		fprintf(stderr, "invalid slot index\n");
		return -1;
	}
	if (!checkeol(arg)) {
		fprintf(stderr, "too many arguments\n");
		return -1;
	}
	return v;
}

int puthex(char *dst, char *src)
{
	int n = 0;
	char buf[3];

	while (*src && *(src + 1)) {
		buf[0] = *src++;
		buf[1] = *src++;
		buf[2] = '\0';
		*dst++ = strtol(buf, NULL, 16);
		n++;
	}
	return n;
}

int getcmd(char *buf, int sz)
{

	if (interactive)
		write(2, "> ", 3);
	while (sz > 1) {
		if (read(0, buf, 1) <= 0)
			return 0;
		if (*buf == '\n') {
			*buf = '\0';
			return 1;
		}
		buf++;
		sz--;
	}
	write(2, "line too long\n", 14);
	drain();
	return -1;
}

int main()
{
	char buf[BUF_SIZE];
	char *cmd, *arg, *dst;
	size_t l;
	int i ;
	long v;

	interactive = isatty(0);

	while ( (l = getcmd(buf, BUF_SIZE)) ) {
		if (l < 0)
			continue;
		if ( !(cmd = strtok(buf, " \t")) )
			continue;
		if (!strcmp(cmd, "malloc")) {
			if ( (i = findfree()) < 0 ) {
				fprintf(stderr, "no more empty slots\n");
				continue;
			}
			if ( !(arg = strtok(NULL, " \t")) ) {
				fprintf(stderr, "missing size argument\n");
				continue;
			}
			v = strtol(arg, &arg, 0);
			if (v < 0) {
				fprintf(stderr, "invalid size\n");
				continue;
			}
			if (!checkeol(arg)) {
				fprintf(stderr, "too many arguments\n");
				continue;
			}
			slot[i] = malloc(v);
			fprintf(stderr, "[%d] %p\n", i, slot[i]);
		} else if (!strcmp(cmd, "free")) {
			if ( (v = getslot()) < 0 )
				continue;
			free(slot[v]);
		} else if (!strcmp(cmd, "list")) {
			for (i = 0; i < MAX_SLOTS; i++) {
				if (slot[i])
					printf("[%d] %p\n", i, slot[i]);
			}
		} else if (!strcmp(cmd, "clear")) {
			if ( (v = getslot()) < 0 )
				continue;
			slot[v] = NULL;
		} else if (!strcmp(cmd, "writestr")) {
			if ( (v = getslot()) < 0 )
				continue;
			if ( !(arg = strtok(NULL, " \t")) ) {
				fprintf(stderr, "missing string\n");
				continue;
			}
			strcpy(slot[v], arg);
		} else if (!strcmp(cmd, "writehex")) {
			if ( (v = getslot()) < 0 )
				continue;
			dst = slot[v];
			while (arg = strtok(NULL, " \t")) {
				if ( (i = puthex(dst, arg)) < 0 ) {
					fprintf(stderr, "syntax error\n");
					continue;
				}
				dst += i;
			}
		} else {
			fprintf(stderr, "unknown command\n");
		}
	}
	printf("\n");
}
